#!/usr/bin/env python
# filename: pl.py


#
# Copyright (c) 2020 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import sys

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import Patch

import seaborn as sns

from natsort import natsorted



def feature_kde(data, x, y, hue=None, hue_order=None, colors=None, thresh=0.1,
                show_scatter=True, scatter_size=5, scatter_alpha=0.2,
                fill=False, kde_fill_alpha=0.7, kde_line_alpha=1.0,
                highlight_index=None, highlight_x=None, highlight_y=None, highlight_marker='x',
                highlight_size=90, highlight_color='k', highlight_name=None, highlight_alpha=0.8,
                xlabel=None, ylabel=None, equal_axes=True,
                legend_loc='best', legend_title=None, show_legend_title=True,
                return_ax=False, figsize=[6, 6], figfile=None):
    '''
    Produces a 2-dimensional KDE plot of two features.

    Args:

        data (pd.DataFramne): A pandas dataframe containing the input data.

        x (str): Name of the column in ``df`` containing the feature to be plotted on the x-axis.

        y (str): Name of the column in ``df`` containing the feature to be plotted on the y-axis.

        hue (str): Name of the column in ``df`` containing categories for hue values. For scatter plots, 
                   the categories in ``hue`` will be plotted as different colored points. For KDE plots,
                   ``hue```` categories will each be plotted as differently colored KDE plots
                   on the same plot. 

        hue_order (iterable): Iterable of hue categories, in the order they should be plotted and listed
                              in the legend. If ```hue_order``` contains only a subset of the categories
                              present in ```data[hue]```, only the categories supplied in ```hue_order```
                              will be plotted.

        colors (iterable): List of colors to be used for ```'hue'``` categories. If ```'colors'``` is
                           shorter than the list of hue categories, colors will be reused.

        thresh (float): Threshold for the KDE. Default is ```0.1```.
        
        show_scatter (bool): Show the scatterplot beneath the transparent KDE plot. Default is ```True```.

        scatter_size (int, float): Size of the scatter points. Default is ```5```.

        scatter_alpha (float): Alpha of the scatter points. Default is ```0.2```.

        fill (bool): Fill the KDE plot. Default is ```True```.

        kde_fill_alpha (float): Alpha for the filled KDE plot. If ```fill``` is ```False```,
                                this option is ignored. Default is ```0.7```.
        
        kde_line_alpha (float): Alpha for the KDE plot lines. Default is ```1.0```.

        highlight_index (iterable): An iterabile of index names (present in ```data```) of points
                                    to be highlighted on the KDE plot. If provided, ```highlight_x```
                                    and ```highlight_y``` are ignored.

        highlight_x (iterable): An iterable of x-values for highlighted points. Also requires
                                ```highlight_y```.
        
        highlight_y (iterable): An iterable of y-values for highlighted points. Also requires
                                ```highlight_x```.
        
        highlight_marker (str): The marker style to be used for highlight points. Accepts 
                                standard matplotlib marker styles. Default is ```'x'```. 

        highlight_size (int): Size of the highlight marker. Default is ```90```.

        highlight_color (string or RGB list): Color of the highlight points. Default is black.

        highlight_name (str): Name of the highlights, to be used in the legend. If not supplied,
                              highlight points will not be included in the legend.
        
        highlight_alpha (float): Alpha for the highlight points. Default is ```0.8```.

        xlabel (str): Label for the x-axis. By default, the value for ```x``` is used.

        ylabel (str): Label for the y-axis. By default, the value for ```y``` is used.

        equal_axes (bool): If ```True```, the the limits of the x- and y-axis will be equal.
                           Default is ```True```.
        
        legend_loc (str): Location for the legend. Uses standard matplotlib locations. Default
                          is ```'best'```.

        legend title (str): Title for the legend. By default, ```hue``` is used.

        show_legend_title (bool): Whether or not to show the legend title. Default is ```True```.

        return_ax (bool): If ```True```, return the plot's ```ax``` object. Will not show or save
                          the plot. Default is ```False```.

        figsize (list): A list containg the dimensions of the plot. Default is ```[6, 6]```.

        figfile (str): Path to which the figure will be saved. If not provided, the figure will be
                       shown but not saved to file.
    '''

    # input data
    data = data.copy()
    if x not in data.columns.values:
        print('"{}" is not a column in the supplied dataframe'.format(x))
        return
    if y not in data.columns.values:
        print('"{}" is not a column in the supplied dataframe'.format(y))
        return

    # hue
    if hue is not None:
        if hue_order is None:
            hue_order = natsorted(list(set(data[hue])))
        data = data[data[hue].isin(hue_order)]
    else:
        hue_order = []

    # colors
    n_colors = max(1, len(hue_order))
    if colors is None:
        colors = sns.color_palette(n_colors=n_colors)
        
    plt.figure(figsize=figsize)

    # scatterplots
    if show_scatter:
        if hue_order:
            for h, c in zip(hue_order, colors):
                d = data[data[hue] == h]
                plt.scatter(d[x], d[y], c=[c], s=scatter_size,
                            alpha=scatter_alpha, linewidths=0)
        else:
            plt.scatter(data[x], data[y], c=[colors[0]], s=scatter_size,
                            alpha=scatter_alpha, linewidths=0)

    # kdeplot
    if fill:
        sns.kdeplot(data=data, x=x, y=y, hue=hue, fill=True, alpha=kde_fill_alpha,
                    hue_order=hue_order, palette=colors, thresh=thresh)
    ax = sns.kdeplot(data=data, x=x, y=y, hue=hue, alpha=kde_line_alpha,
                     hue_order=hue_order, palette=colors, thresh=thresh)
    
    # highlighted points
    highlight = any([highlight_index is not None, all([highlight_x is not None, highlight_y is not None])])
    if highlight:
        if highlight_index is not None:
            hidata = data.loc[highlight_index]
            highlight_x = hidata[x]
            highlight_y = hidata[y]
        plt.scatter(highlight_x, highlight_y, zorder=10,
                    s=highlight_size,
                    c=highlight_color,
                    alpha=highlight_alpha,
                    marker=highlight_marker)

    # legend
    legend_labels = hue_order
    if fill:
        handles = []
        for c in colors:
            f = Patch(fc=c, alpha=kde_fill_alpha / 3)
            e = Patch(ec=c, fill=False, lw=1.5)
            handles.append((f, e))
#         handles = [Patch(fc=c, ec=c, alpha=kde_fill_alpha / 3, label=h) for c, h in zip(colors, hue_order)]
    else:
        handles = [Line2D([0], [0], color=c) for c in colors]
    if highlight_name is not None:
        legend_labels.append(highlight_name)
        handles.append(Line2D([0], [0], marker=highlight_marker, color='w',
                              mec=highlight_color,
                              mfc=highlight_color,
                              ms=highlight_size / 10))
    if show_legend_title:
        legend_title = legend_title if legend_title is not None else hue
    else:
        legend_title = None
    ax.legend(handles, legend_labels, loc=legend_loc, fontsize=13, title=legend_title, frameon=False)
    
    # style the plot
    ax.set_xlabel(xlabel if xlabel is not None else x, fontsize=16)
    ax.set_ylabel(ylabel if ylabel is not None else y, fontsize=16)
    ax.tick_params(axis='both', labelsize=13)

    for spine in ['right', 'top']:
        ax.spines[spine].set_visible(False)
    for spine in ['left', 'bottom']:
        ax.spines[spine].set_position(('outward', 10))

    if equal_axes:
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        axlim = [min([xlim[0], ylim[0]]), max([xlim[1], ylim[1]])]
        ax.set_xlim(axlim)
        ax.set_ylim(axlim)
    
    if return_ax:
        return ax
    elif figfile is not None:
        plt.tight_layout()
        plt.savefig(figfile)
    else:
        plt.show()


    
def feature_scatter(data, x, y, hue=None, hue_order=None, colors=None, marker='o', size=20, alpha=0.6,
                    highlight_index=None, highlight_x=None, highlight_y=None, highlight_marker='o',
                    highlight_size=90, highlight_color='k', highlight_name=None, highlight_alpha=0.9,
                    xlabel=None, ylabel=None, equal_axes=True,
                    legend_loc='best', legend_title=None, show_legend_title=True,
                    return_ax=False, figsize=[6, 6], figfile=None):
    '''
    Docstring for feature_scatter.
    '''
    # input data
    data = data.copy()
    if x not in data.columns.values:
        print('"{}" is not a column in the supplied dataframe'.format(x))
        return
    if y not in data.columns.values:
        print('"{}" is not a column in the supplied dataframe'.format(y))
        return
    
    # hue
    if hue is not None:
        if hue_order is None:
            hue_order = natsorted(list(set(data[hue])))
        data = data[data[hue].isin(hue_order)]
    else:
        hue_order = []

    # colors
    n_colors = max(1, len(hue_order))
    if colors is None:
        colors = sns.color_palette(n_colors=n_colors)
        
    plt.figure(figsize=figsize)
    ax = plt.gca()

    # scatterplot
    if hue_order:
        for h, c in zip(hue_order, colors):
            d = data[data[hue] == h]
            plt.scatter(d[x], d[y], c=[c], s=size, marker=marker,
                        alpha=alpha, linewidths=0, label=h)
    else:
        plt.scatter(data[x], data[y], c=[colors[0]], s=size, marker=marker,
                        alpha=alpha, linewidths=0)

    # highlighted points
    highlight = any([highlight_index is not None, all([highlight_x is not None, highlight_y is not None])])
    if highlight:
        if highlight_index is not None:
            hidata = data.loc[highlight_index]
            highlight_x = hidata[x]
            highlight_y = hidata[y]
        plt.scatter(highlight_x, highlight_y, zorder=10,
                    s=highlight_size,
                    c=highlight_color,
                    alpha=highlight_alpha,
                    marker=highlight_marker,
                    label=highlight_name)
    # legend
    ax.legend(loc=legend_loc, fontsize=13, title=legend_title, frameon=False)
    
    # style the plot
    ax.set_xlabel(xlabel if xlabel is not None else x, fontsize=16)
    ax.set_ylabel(ylabel if ylabel is not None else y, fontsize=16)
    ax.tick_params(axis='both', labelsize=13)

    for spine in ['right', 'top']:
        ax.spines[spine].set_visible(False)
    for spine in ['left', 'bottom']:
        ax.spines[spine].set_position(('outward', 10))

    if equal_axes:
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        axlim = [min([xlim[0], ylim[0]]), max([xlim[1], ylim[1]])]
        ax.set_xlim(axlim)
        ax.set_ylim(axlim)
    
    if return_ax:
        return ax
    elif figfile is not None:
        plt.tight_layout()
        plt.savefig(figfile)
    else:
        plt.show()


def cellhash_ridge(data, hashname, category, colors=None, alpha=1.0,
                   categories=None, hide_extra_categories=False, rename=None, xmax=14,
                   ylabel_fontsize=11, xlabel=None, xlabel_fontsize=12,
                   feature_label_xoffset=5, figfile=None):
    '''
    Docstring for feature_ridge.
    '''
    
    # input data
    data = data.copy()
    if category not in data.columns.values:
        print('"{}" is not a column in the supplied dataframe'.format(category))
        return
    
    # rename
    if rename is None:
        rename = {}
    else:
        if not any([k in data.columns.values for k in rename.keys()]):
            rename = {v: k for k, v in rename.items()}

    # categories
    category_set = list(set(data[category]))
    if categories is None:
        feature_cats = natsorted([c for c in category_set if rename.get(c, c) in data.columns.values])
        extra_cats = natsorted([c for c in category_set if rename.get(c, c) not in feature_cats])
        categories = feature_cats + extra_cats
    else:
        feature_cats = categories
        if hide_extra_categories:
            extra_cats = []
        else:
            extra_cats = [c for c in category_set if rename.get(c, c) not in feature_cats]
        categories = feature_cats + extra_cats

    # colors
    if colors is None:
        n_colors = len(feature_cats)
        colors = list(sns.color_palette(n_colors=n_colors))
        n_greys = len(extra_cats)
        greys = list(plt.get_cmap('Greys')(np.linspace(0, 1, n_greys + 2))[1:-1, :3])
        cdict = {h: c for h, c in zip(categories, colors + greys)}
    elif isinstance(colors, (list, tuple, np.ndarray, pd.core.series.Series)):
        colors = list(colors)
        if len(colors) < len(categories):
            n_greys = len(categories) - len(colors)
            greys = list(plt.get_cmap('Greys')(np.linspace(0, 1, n_greys + 2))[1:-1, :3])
        cdict = {h: c for h, c in zip(categories, colors + greys)}
    else:
        cdict = colors
        if len(cdict) < len(categories):
            missing = [k for k in categories if k not in cdict]
            n_greys = len(missing)
            greys = list(plt.get_cmap('Greys')(np.linspace(0, 1, n_greys + 2))[1:-1, :3])
            for m, g in zip(missing, greys):
                cdict[m] = g
    colors = [cdict[c] for c in categories]

    # plot
    g = sns.FacetGrid(data, row=category, hue=category,
                      aspect=7.5, height=0.75, palette=colors,
                      row_order=categories, hue_order=categories)
    g.map(sns.kdeplot, hashname, clip_on=False, shade=True, alpha=alpha, lw=1.5)
    g.map(sns.kdeplot, hashname, clip_on=False, color="w", lw=3)
    g.map(plt.axhline, y=0, lw=2, clip_on=False)

    def label(x, color, label):
        ax = plt.gca()
        ax.text(0, .2, label, fontweight="bold", color=color,
                fontsize=ylabel_fontsize,
                ha="left", va="center", transform=ax.transAxes)
    g.map(label, hashname)

    # set the subplots to overlap
    g.fig.subplots_adjust(hspace=0.1)

    # remove axes details that don't play well with overlap
    g.set_titles("")
    g.set(xticks=range(0, xmax + 1, 2))
    g.set(yticks=[])
    g.set(xlim=[-feature_label_xoffset, xmax + 0.25])
    g.despine(bottom=True, left=True)
    
    # xlabel
    if xlabel is not None:
        g.set(xlabel=xlabel)
    
    xlabel_position = ((xmax / 2) + feature_label_xoffset) / (xmax + feature_label_xoffset)
    for ax in g.axes.flat:
        ax.set_xlabel(ax.get_xlabel(),
                      x=xlabel_position,
                      fontsize=xlabel_fontsize)
        
    for ax in g.axes.flat:
        ax.set_xlabel(ax.get_xlabel(), fontsize=xlabel_fontsize)

    if figfile is not None:
        g.savefig(figfile)
    else:
        plt.show()


def feature_ridge(data, features, colors=None, rename=None,
                  xlabel='UMI count ($\mathregular{log_2}$)', 
                  ylabel_fontsize=11, xlabel_fontsize=12,
                  feature_label_xoffset=5, xmax=14, alpha=1.0,
                  figfile=None):
    '''
    Docstring for feature_ridge.
    '''
    
    # input data
    data = data.copy()
    features = [f for f in features if f in data.columns.values]
    melted = data.melt(value_vars=features, var_name='feature')
    
    # rename
    if rename is None:
        rename = {}
    else:
        if not any([k in data.columns.values for k in rename.keys()]):
            rename = {v: k for k, v in rename.items()}

    # colors
    if colors is None:
        n_colors = len(features)
        colors = list(sns.color_palette(n_colors=n_colors))
        cdict = {h: c for h, c in zip(features, colors)}
    elif isinstance(colors, (list, tuple, np.ndarray, pd.core.series.Series)):
        cdict = {h: c for h, c in zip(features, colors)}
    else:
        cdict = colors
    colors = [cdict[f] for f in features]

    # plot
    g = sns.FacetGrid(melted, row='feature', hue='feature',
                      aspect=7.5, height=0.75, palette=colors,
                      row_order=features, hue_order=features)
    g.map(sns.kdeplot, 'value', clip_on=False, shade=True, alpha=alpha, lw=1.5)
    g.map(sns.kdeplot, 'value', clip_on=False, color="w", lw=3)
    g.map(plt.axhline, y=0, lw=2, clip_on=False)

    def label(x, color, label):
        ax = plt.gca()
        ax.text(0, .2, label, fontweight="bold", color=color,
                fontsize=ylabel_fontsize,
                ha="left", va="center", transform=ax.transAxes)
    g.map(label, 'feature')

    # set the subplots to overlap
    g.fig.subplots_adjust(hspace=0.1)

    # remove axes details that don't play well with overlap
    g.set_titles("")
    g.set(xticks=range(0, xmax + 1, 2))
    g.set(yticks=[])
    g.set(xlim=[-feature_label_xoffset, xmax + 0.25])
    g.despine(bottom=True, left=True)
    
    # xlabel
    g.set(xlabel=xlabel)
    xlabel_position = ((xmax / 2) + feature_label_xoffset) / (xmax + feature_label_xoffset)
    for ax in g.axes.flat:
        ax.set_xlabel(ax.get_xlabel(),
                      x=xlabel_position,
                      fontsize=xlabel_fontsize)

    if figfile is not None:
        g.savefig(figfile)
    else:
        plt.show()






#!/usr/bin/env python
# filename: pair.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import csv
import sys
import traceback

from Bio.Seq import Seq

from .sequence import Sequence
from ..utils import germlines
from ..utils.alignment import global_alignment
from ..utils.utilities import nested_dict_lookup


if sys.version_info[0] > 2:
    STR_TYPES = [str, ]
else:
    STR_TYPES = [str, unicode]


class Pair(object):
    '''
    Holds a pair of sequences, corresponding to HC and LC of a single mAb.

    Input is a list of dicts, with each dict containing sequence information from a single
    chain, formatted as would be returned from a query on a MongoDB database containing
    AbStar output.
    '''
    def __init__(self, seqs, name=None, chain_selection_func=None):
        self._seqs = seqs
        self._receptor = None
        self._heavy = None
        self._light = None
        self._alpha = None
        self._beta = None
        self._delta = None
        self._gamma = None
        self._heavies = None
        self._lights = None
        self._alphas = None
        self._betas = None
        self._deltas = None
        self._gammas = None
        self._name = name
        self._fasta = None
        self._sample = None
        self._subject = None
        self._group = None
        self._experiment = None
        self._timepoint = None
        self._is_pair = None
        self._lineage = None
        self._select_chain = chain_selection_func if chain_selection_func is not None else self._chain_selector


    def __eq__(self, other):
        return (self.heavy, self.light) == (other.heavy, other.light)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((self.heavy, self.light))
        

    @property
    def receptor(self):
        if self._receptor is None:
            if all([s['chain'] is not None for s in self._seqs]):
                if all([s['chain'] in ['heavy', 'kappa', 'lambda'] for s in self._seqs]):
                    self._receptor = 'bcr'
                elif all([s['chain'] in ['alpha', 'beta', 'delta', 'gamma'] for s in self._seqs]):
                    self._receptor = 'tcr'
                else:
                    self._receptor = 'unknown'
            elif all([s['locus'] is not None for s in self._seqs]):
                if all([s['locus'].lower() in ['igh', 'igk', 'igl'] for s in self._seqs]):
                    self._receptor = 'bcr'
                elif all([s['locus'].lower() in ['tra', 'trb', 'trd', 'trg'] for s in self._seqs]):
                    self._receptor = 'tcr'
                else:
                    self._receptor = 'unknown'
        return self._receptor

    @receptor.setter
    def receptor(self, receptor):
        self._receptor = receptor


    @property
    def heavy(self):
        if self._heavy is None:
            self._heavy = self._select_chain(self.heavies)
        return self._heavy

    @heavy.setter
    def heavy(self, heavy):
        self._heavy = heavy

    @property
    def light(self):
        if self._light is None:
            self._light = self._select_chain(self.lights)
        return self._light

    @light.setter
    def light(self, light):
        self._light = light


    @property
    def alpha(self):
        if self._alpha is None:
            self._alpha = self._select_chain(self.alphas)
        return self._alpha

    @alpha.setter
    def alpha(self, alpha):
        self._alpha = alpha

    
    @property
    def beta(self):
        if self._beta is None:
            self._beta = self._select_chain(self.betas)
        return self._beta

    @beta.setter
    def beta(self, beta):
        self._beta = beta

    
    @property
    def delta(self):
        if self._delta is None:
            self._delta = self._select_chain(self.deltas)
        return self._delta

    @delta.setter
    def delta(self, delta):
        self._delta = delta


    @property
    def gamma(self):
        if self._gamma is None:
            self._gamma = self._select_chain(self.gammas)
        return self._gamma

    @gamma.setter
    def gamma(self, gamma):
        self._gamma = gamma

    
    @property
    def heavies(self):
        if self._heavies is None:
            if all([s['chain'] is not None for s in self._seqs]):
                self._heavies = [s for s in self._seqs if s['chain'] == 'heavy']
            elif all([s['locus'] is not None for s in self._seqs]):
                self._heavies = [s for s in self._seqs if s['locus'] == 'IGH']
        return self._heavies
    
    @property
    def lights(self):
        if self._lights is None:
            if all([s['chain'] is not None for s in self._seqs]):
                self._lights = [s for s in self._seqs if s['chain'] in ['kappa', 'lambda']]
            elif all([s['locus'] is not None for s in self._seqs]):
                self._lights = [s for s in self._seqs if s['locus'] in ['IGK', 'IGL']]
        return self._lights
    
    @property
    def alphas(self):
        if self._alphas is None:
            if all([s['chain'] is not None for s in self._seqs]):
                self._alphas = [s for s in self._seqs if s['chain'] == 'alpha']
            elif all([s['locus'] is not None for s in self._seqs]):
                self._alphas = [s for s in self._seqs if s['locus'] == 'TRA']
        return self._alphas

    @property
    def betas(self):
        if self._betas is None:
            if all([s['chain'] is not None for s in self._seqs]):
                self._betas = [s for s in self._seqs if s['chain'] == 'beta']
            elif all([s['locus'] is not None for s in self._seqs]):
                self._betas = [s for s in self._seqs if s['locus'] == 'TRB']
        return self._betas

    @property
    def deltas(self):
        if self._deltas is None:
            if all([s['chain'] is not None for s in self._seqs]):
                self._deltas = [s for s in self._seqs if s['chain'] == 'delta']
            elif all([s['locus'] is not None for s in self._seqs]):
                self._deltas = [s for s in self._seqs if s['locus'] == 'TRD']
        return self._deltas

    @property
    def gammas(self):
        if self._gammas is None:
            if all([s['chain'] is not None for s in self._seqs]):
                self._gammas = [s for s in self._seqs if s['chain'] == 'gamma']
            elif all([s['locus'] is not None for s in self._seqs]):
                self._gammas = [s for s in self._seqs if s['locus'] == 'TRG']
        return self._gammas

    @property
    def is_pair(self):
        if all([self.heavy is not None, self.light is not None]):
            return True
        elif all([self.alpha is not None, self.beta is not None]):
            return True
        elif all([self.gamma is not None, self.delta is not None]):
            return True
        return False

    @property
    def lineage(self):
        if self._lineage is None:
            if 'clonify' in self.heavy:
                self._lineage = self.heavy['clonify']['id']
        return self._lineage

    # @property
    # def vrc01_like(self):
    #     if self._vrc01_like is None:
    #         if any([self.heavy is None, self.light is None]):
    #             self._vrc01_like = False
    #         else:
    #             try:
    #                 # abstar's JSON output format
    #                 self._vrc01_like = all([self.heavy['v_gene']['gene'] == 'IGHV1-2', self.light['cdr3_len'] == 5])
    #             except KeyError:
    #                 try:
    #                     # AIRR format
    #                     self._vrc01_like = all([self.heavy['v_call'] == 'IGHV1-2', self.light['junction_aa_length'] == 7])
    #                 except KeyError:
    #                     # unknown format
    #                     pass
    #     return self._vrc01_like

    # @property
    # def loose_vrc01_like(self):
    #     if self._loose_vrc01_like is None:
    #         try:
    #             # abstar's JSON output format
    #             if any([l['cdr3_len'] == 5 for l in p._lights]) and any([h['v_gene']['gene'] == 'IGHV1-2' for h in p._heavies]):
    #                 self._loose_vrc01_like = True
    #             else:
    #                 self._loose_vrc01_like = False
    #         except KeyError:
    #             try:
    #                 # AIRR format
    #                 if any([l['junction_aa_length'] == 7 for l in p._lights]) and any([h['v_call'] == 'IGHV1-2' for h in p._heavies]):
    #                     self._loose_vrc01_like = True
    #                 else:
    #                     self._loose_vrc01_like = False
    #             except:
    #                 # unknown format
    #                 pass
    #     return self._loose_vrc01_like

    @property
    def name(self):
        if self._name is None:
            if self.heavy is not None:
                try:
                    # abstar's JSON format
                    self._name = self.heavy['seq_id']
                except KeyError:
                    try:
                        # AIRR format
                        self._name = self.heavy['sequence_id']
                    except KeyError:
                        # unknown format
                        pass
            elif self.light is not None:
                try:
                    # abstar's JSON format
                    self._name = self.light['seq_id']
                except KeyError:
                    try:
                        # AIRR format
                        self._name = self.light['sequence_id']
                    except KeyError:
                        # unknown format
                        pass
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    # @property
    # def sample(self):
    #     if self._sample is None:
    #         slist = []
    #         if self.experiment is not None:
    #             slist.append(str(self.experiment))
    #         if self.group is not None:
    #             slist.append(str(self.group))
    #         if self.subject is not None:
    #             slist.append(str(self.subject))
    #         if self.timepoint is not None:
    #             slist.append(str(self.timepoint))
    #         if slist:
    #             self._sample = '|'.join(slist)
    #     return self._sample

    # @property
    # def subject(self):
    #     if self._subject is None:
    #         if self.heavy is not None and 'subject' in self.heavy.keys():
    #             self._subject = self.heavy['subject']
    #         elif self.light is not None and 'subject' in self.light.keys():
    #             self._subject = self.light['subject']
    #     return self._subject

    # @subject.setter
    # def subject(self, subject):
    #     self._subject = subject

    # @property
    # def group(self):
    #     if self._group is None:
    #         if self.heavy is not None and 'group' in self.heavy.keys():
    #             self._group = self.heavy['group']
    #         elif self.light is not None and 'group' in self.light.keys():
    #             self._group = self.light['group']
    #     return self._group

    # @group.setter
    # def group(self, group):
    #     self._group = group

    # @property
    # def experiment(self):
    #     if self._experiment is None:
    #         if self.heavy is not None and 'experiment' in self.heavy.keys():
    #             self._experiment = self.heavy['experiment']
    #         elif self.light is not None and 'experiment' in self.light.keys():
    #             self._experiment = self.light['experiment']
    #     return self._experiment

    # @experiment.setter
    # def experiment(self, experiment):
    #     self._experiment = experiment

    # @property
    # def timepoint(self):
    #     if self._timepoint is None:
    #         if self.heavy is not None and 'timepoint' in self.heavy.keys():
    #             self._timepoint = self.heavy['timepoint']
    #         elif self.light is not None and 'timepoint' in self.light.keys():
    #             self._timepoint = self.light['timepoint']
    #     return self._timepoint

    # @timepoint.setter
    # def timepoint(self, timepoint):
    #     self._timepoint = timepoint


    # def refine(self, heavy=True, light=True, species='human'):
    #     for seq in [s for s in [self.heavy, self.light] if s is not None]:
    #         try:
    #             self.remove_ambigs(seq)
    #             self._refine_v(seq, species)
    #             self._refine_j(seq, species)
    #             self._retranslate(seq)
    #         except:
    #             print('REFINEMENT FAILED: {}, {} chain'.format(seq['seq_id'], seq['chain']))
    #             print(traceback.format_exception_only(*sys.exc_info()[:2]))


    # @staticmethod
    # def remove_ambigs(seq):
    #     # fix Ns in the nucleotide sequence
    #     vdj = ''
    #     for s, g in zip(seq['vdj_nt'], seq['vdj_germ_nt']):
    #         if s.upper() == 'N':
    #             vdj += g
    #         else:
    #             vdj += s
    #     seq['vdj_nt'] = vdj
    #     # fix Xs in the amino acid sequence
    #     vdj = ''
    #     for s, g in zip(seq['vdj_aa'], seq['vdj_germ_aa']):
    #         if s.upper() == 'X':
    #             vdj += g
    #         else:
    #             vdj += s
    #     seq['vdj_aa'] = vdj

    # @staticmethod
    # def _refine_v(seq, species):
    #     '''
    #     Completes the 5' end of a a truncated sequence with germline nucleotides.
    #     Input is a MongoDB dict (seq) and the species.
    #     '''
    #     vgerm = germlines.get_germline(seq['v_gene']['full'], species)
    #     aln = global_alignment(seq['vdj_nt'], vgerm)
    #     prepend = ''
    #     for s, g in zip(aln.aligned_query, aln.aligned_target):
    #         if s != '-':
    #             break
    #         else:
    #             prepend += g
    #     seq['vdj_nt'] = prepend + seq['vdj_nt']

    # @staticmethod
    # def _refine_j(seq, species):
    #     '''
    #     Completes the 3' end of a a truncated sequence with germline nucleotides.
    #     Input is a MongoDB dict (seq) and the species.
    #     '''
    #     jgerm = germlines.get_germline(seq['j_gene']['full'], species)
    #     aln = global_alignment(seq['vdj_nt'], jgerm)
    #     append = ''
    #     for s, g in zip(aln.aligned_query[::-1], aln.aligned_target[::-1]):
    #         if s != '-':
    #             break
    #         else:
    #             append += g
    #     seq['vdj_nt'] = seq['vdj_nt'] + append[::-1]

    # @staticmethod
    # def _retranslate(seq):
    #     '''
    #     Retranslates a nucleotide sequence following refinement.
    #     Input is a Pair sequence (basically a dict of MongoDB output).
    #     '''
    #     if len(seq['vdj_nt']) % 3 != 0:
    #         trunc = len(seq['vdj_nt']) % 3
    #         seq['vdj_nt'] = seq['vdj_nt'][:-trunc]
    #     seq['vdj_aa'] = Seq(seq['vdj_nt']).translate()


    def fasta(self, name_field='sequence_id', sequence_field='sequence', append_chain=True):
        '''
        Returns the sequence pair as a fasta string. If the Pair object contains
        both heavy and light chain sequences, both will be returned as a single string.

        By default, the fasta string contains the 'vdj_nt' sequence for each chain. To change,
        use the <key> option to select an alternate sequence.

        By default, the chain (heavy or light) will be appended to the sequence name:

        >MySequence_heavy

        To just use the pair name (which will result in duplicate sequence names for Pair objects
        with both heavy and light chains), set <append_chain> to False.
        '''
        fastas = []
        for s, chain in [(self.heavy, 'heavy'), (self.light, 'light')]:
            if s is not None:
                c = '_{}'.format(chain) if append_chain else ''
                fastas.append('>{}{}\n{}'.format(s[name_field], c, s[sequence_field]))
        return '\n'.join(fastas)


    def summarize(self, annotation_format='airr'):
        try:
            if annotation_format == 'airr':
                v_key = 'v_gene'
                d_key = 'd_gene'
                j_key = 'j_gene'
                junc_key = 'junction_aa'
                vident_key = 'v_identity'
                isotype_key = 'isotype'
            elif annotation_format == 'json':
                v_key = 'v_gene.gene'
                d_key = 'd_gene.gene'
                j_key = 'j_gene.gene'
                junc_key = 'junc_aa'
                vident_key = 'nt_identity.v'
                isotype_key = 'isotype'
            vline = ''
            dline = ''
            jline = ''
            juncline = ''
            identline = ''
            isotypeline = ''
            if self.heavy is not None:
                vline += nested_dict_lookup(self.heavy, v_key.split('.'))
                dline += nested_dict_lookup(self.heavy, d_key.split('.'))
                jline += nested_dict_lookup(self.heavy, j_key.split('.'))
                juncline += nested_dict_lookup(self.heavy, junc_key.split('.'))
                identline += nested_dict_lookup(self.heavy, vident_key.split('.'))
                isotypeline += nested_dict_lookup(self.heavy, isotype_key.split('.'))
            pad = max([len(l) for l in [vline, dline, jline, juncline, identline, isotypeline]]) + 4
            vline += ' ' * (pad - len(vline))
            dline += ' ' * (pad - len(dline))
            jline += ' ' * (pad - len(jline))
            juncline += ' ' * (pad - len(juncline))
            identline += ' ' * (pad - len(identline))
            isotypeline += ' ' * (pad - len(isotypeline))
            if self.light is not None:
                vline += nested_dict_lookup(self.light, v_key.split('.'))
                jline += nested_dict_lookup(self.light, j_key.split('.'))
                juncline += nested_dict_lookup(self.light, junc_key.split('.'))
                identline += nested_dict_lookup(self.light, vident_key.split('.'))
            
            header_len = max([len(l) for l in [vline, dline, jline, juncline, identline, isotypeline]])
            header_spaces = int((header_len - len(self.name)) / 2)
            print(f"{' ' * header_spaces}{self.name}")
            print('-' * header_len)
            for l in [vline, dline, jline, juncline, identline, isotypeline]:
                print(l)
            print('')
        except:
            return
        


    def _chain_selector(self, seqs):
        if len(seqs) == 0:
            return None
        if all(['umis' in s for s in seqs]):
            sorted_seqs = sorted(seqs, key=lambda x: x['umis'], reverse=True)
            return sorted_seqs[0]
        else:
            return seqs[0]







def get_pairs(db, collection, experiment=None, subject=None, group=None, name='seq_id',
    delim=None, delim_occurance=1, pairs_only=False, chain_selection_func=None):
    '''
    Gets sequences and assigns them to the appropriate mAb pair, based on the sequence name.

    Inputs:

    ::db:: is a pymongo database connection object
    ::collection:: is the collection name, as a string
    If ::subject:: is provided, only sequences with a 'subject' field matching ::subject:: will
        be included. ::subject:: can be either a single subject (as a string) or an iterable
        (list or tuple) of subject strings.
    If ::group:: is provided, only sequences with a 'group' field matching ::group:: will
        be included. ::group:: can be either a single group (as a string) or an iterable
        (list or tuple) of group strings.
    ::name:: is the dict key of the field to be used to group the sequences into pairs.
        Default is 'seq_id'
    ::delim:: is an optional delimiter used to truncate the contents of the ::name:: field.
        Default is None, which results in no name truncation.
    ::delim_occurance:: is the occurance of the delimiter at which to trim. Trimming is performed
        as delim.join(name.split(delim)[:delim_occurance]), so setting delim_occurance to -1 will
        trucate after the last occurance of delim. Default is 1.
    ::pairs_only:: setting to True results in only truly paired sequences (pair.is_pair == True)
        will be returned. Default is False.

    Returns a list of Pair objects, one for each mAb pair.
    '''
    match = {}
    if subject is not None:
        if type(subject) in (list, tuple):
            match['subject'] = {'$in': subject}
        elif type(subject) in STR_TYPES:
            match['subject'] = subject
    if group is not None:
        if type(group) in (list, tuple):
            match['group'] = {'$in': group}
        elif type(group) in STR_TYPES:
            match['group'] = group
    if experiment is not None:
        if type(experiment) in (list, tuple):
            match['experiment'] = {'$in': experiment}
        elif type(experiment) in STR_TYPES:
            match['experiment'] = experiment
    seqs = list(db[collection].find(match))
    return assign_pairs(seqs, name=name, delim=delim,
        delim_occurance=delim_occurance, pairs_only=pairs_only,
        chain_selection_func=chain_selection_func)


def assign_pairs(seqs, id_key='sequence_id', delim=None, delim_occurance=1, pairs_only=False,
                 chain_selection_func=None, tenx_annot_file=None):
    '''
    Assigns sequences to the appropriate mAb pair, based on the sequence name.

    Inputs:

    ::seqs:: is a list of dicts, of the format returned by querying a MongoDB containing
        Abstar output.
    ::name:: is the dict key of the field to be used to group the sequences into pairs.
        Default is 'seq_id'
    ::delim:: is an optional delimiter used to truncate the contents of the ::name:: field.
        Default is None, which results in no name truncation.
    ::delim_occurance:: is the occurance of the delimiter at which to trim. Trimming is performed
        as delim.join(name.split(delim)[:delim_occurance]), so setting delim_occurance to -1 will
        trucate after the last occurance of delim. Default is 1.
    ::pairs_only:: setting to True results in only truly paired sequences (pair.is_pair == True)
        will be returned. Default is False.

    Returns a list of Pair objects, one for each mAb pair.
    '''
    # add UMIs to the sequence annotations if a 10xG annotations file is provided
    if tenx_annot_file is not None:
        annots = {}
        umis = {}
        with open(tenx_annot_file) as f:
            reader = csv.DictReader(f)
            for r in reader:
                key = 'consensus_id' if 'consensus_id' in r else 'contig_id'
                umis[r[key]] = r['umis']
                annots[r[key]] = r
        for s in seqs:
            s['umis'] = umis.get(s[id_key], 0)
            s['tenx_annots'] = annots.get(s[id_key], {})

    # identify pairs
    pdict = {}
    for s in seqs:
        if delim is not None:
            pname = delim.join(s[id_key].split(delim)[:delim_occurance])
        else:
            pname = s[id_key]
        if pname not in pdict:
            pdict[pname] = [s, ]
        else:
            pdict[pname].append(s)
    pairs = [Pair(p,
                  name=n,
                  chain_selection_func=chain_selection_func) for n, p in pdict.items()]
    if pairs_only:
        pairs = [p for p in pairs if p.is_pair]
    return pairs


# def deduplicate(pairs, aa=False, ignore_primer_regions=False):
#     '''
#     Removes duplicate sequences from a list of Pair objects.

#     If a Pair has heavy and light chains, both chains must identically match heavy and light chains
#     from another Pair to be considered a duplicate. If a Pair has only a single chain,
#     identical matches to that chain will cause the single chain Pair to be considered a duplicate,
#     even if the comparison Pair has both chains.

#     Note that identical sequences are identified by simple string comparison, so sequences of
#     different length that are identical over the entirety of the shorter sequence are not
#     considered duplicates.

#     By default, comparison is made on the nucleotide sequence. To use the amino acid sequence instead,
#     set aa=True.
#     '''
#     nr_pairs = []
#     just_pairs = [p for p in pairs if p.is_pair]
#     single_chains = [p for p in pairs if not p.is_pair]
#     _pairs = just_pairs + single_chains
#     for p in _pairs:
#         duplicates = []
#         for nr in nr_pairs:
#             identical = True
#             vdj = 'vdj_aa' if aa else 'vdj_nt'
#             offset = 4 if aa else 12
#             if p.heavy is not None:
#                 if nr.heavy is None:
#                     identical = False
#                 else:
#                     heavy = p.heavy[vdj][offset:-offset] if ignore_primer_regions else p.heavy[vdj]
#                     nr_heavy = nr.heavy[vdj][offset:-offset] if ignore_primer_regions else nr.heavy[vdj]
#                     if heavy != nr_heavy:
#                         identical = False
#             if p.light is not None:
#                 if nr.light is None:
#                     identical = False
#                 else:
#                     light = p.light[vdj][offset:-offset] if ignore_primer_regions else p.light[vdj]
#                     nr_light = nr.light[vdj][offset:-offset] if ignore_primer_regions else nr.light[vdj]
#                     if light != nr_light:
#                         identical = False
#             duplicates.append(identical)
#         if any(duplicates):
#             continue
#         else:
#             nr_pairs.append(p)
#     return nr_pairs


# def refine(pairs, heavy=True, light=True, species='human'):
#     refined_pairs = copy.deepcopy(pairs)
#     for p in refined_pairs:
#         p.refine(heavy, light, species)
#     return refined_pairs




def vrc01_like(pair, loose=False):
    if loose:
        return loose_vrc01_like(pair)
    else:
        return strict_vrc01_like(pair)


def strict_vrc01_like(pair):
    if any([pair.heavy is None, pair.light is None]):
        vrc01_like = False
    else:
        try:
            # abstar's JSON output format
            vrc01_like = all([pair.heavy['v_gene']['gene'] == 'IGHV1-2', pair.light['cdr3_len'] == 5])
        except KeyError:
            try:
                # AIRR format
                vrc01_like = all([pair.heavy['v_call'] == 'IGHV1-2', pair.light['junction_aa_length'] == 7])
            except KeyError:
                # unknown format
                vrc01_like = None
    return vrc01_like


def loose_vrc01_like(pair):
    try:
        # abstar's JSON output format
        if any([l['cdr3_len'] == 5 for l in pair.lights]) and any([h['v_gene']['gene'] == 'IGHV1-2' for h in pair.heavies]):
            loose_vrc01_like = True
        else:
            loose_vrc01_like = False
    except KeyError:
        try:
            # AIRR format
            if any([l['junction_aa_length'] == 7 for l in pair.lights]) and any([h['v_call'] == 'IGHV1-2' for h in pair.heavies]):
                loose_vrc01_like = True
            else:
                loose_vrc01_like = False
        except:
            # unknown format
            loose_vrc01_like = None
    return loose_vrc01_like














#!/usr/bin/env python
# filename: alignment.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

from contextlib import contextmanager
import multiprocessing as mp
import time

from . import progbar


def monitor_mp_jobs(results, start_time=None, completion_string='\n', print_progress=True):
    finished = 0
    jobs = len(results)
    while finished < jobs:
        time.sleep(1)
        ready = [ar for ar in results if ar.ready()]
        finished = len(ready)
        if print_progress:
            progbar.progress_bar(finished, jobs, start_time=start_time)
    if print_progress:
        progbar.progress_bar(finished, jobs, start_time=start_time,
                             complete=True, completion_string=completion_string)


def monitor_celery_jobs(results, start_time=None, completion_string='\n', print_progress=True):
    finished = 0
    jobs = len(results)
    while finished < jobs:
        time.sleep(1)
        succeeded = [ar for ar in results if ar.successful()]
        failed = [ar for ar in results if ar.failed()]
        finished = len(succeeded) + len(failed)
        if print_progress:
            progbar.progress_bar(finished, jobs, start_time=start_time)
    if print_progress:
        progbar.progress_bar(finished, jobs, start_time=start_time,
                             complete=True, completion_string=completion_string)


@contextmanager
def poolcontext(*args, **kwargs):
    pool = mp.Pool(*args, **kwargs)
    yield pool
    pool.terminate()



#!/usr/bin/env python
# filename: summary.py


#
# Copyright (c) 2017 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns

from abstar.core.germline import get_germlines


__all__ = ['germline_use_plot', 'cdr3_length_plot', 'vj_heatmap']


#==============================
#           PLOTS
#==============================


def germline_use_plot(seqs, fig_file=None, level='gene', gene='V', species='human', chain='heavy'):
    from .base import barplot
    seqs = [s for s in seqs if s['chain'] == chain]
    germs = get_germlines(species, gene, chain=chain)
    if level == 'fam':
        keys = list(set([g.name.split('-')[0] for g in germs]))
        size = (6, 4)
    elif level == 'gene':
        keys = list(set([g.name.split('*')[0] for g in germs]))
        if gene == 'J':
            size = (6, 4)
        elif gene == 'D':
            size = (8, 4)
        else:
            size = (10, 4)
    gtype = '{}_gene'.format(gene.lower())
    short_chain = chain[0].upper()
    data = [s[gtype][level].rstrip('D') for s in seqs if gtype in s]
    x, y = _aggregate(data, keys=keys)
    x = [i.replace('IG{}{}'.format(short_chain, gene), '{}{}'.format(gene, short_chain)) for i in x]
    colors = _get_germline_plot_colors(x)
    barplot(x, y,
            colors,
            fig_file=fig_file,
            ylabel='Frequency (%)',
            rotate_xtick_labels=True,
            size=size)


def cdr3_length_plot(seqs, fig_file=None, max_len=40, chain='heavy', color=None):
    from .base import barplot
    if chain == 'light':
        chain = ['kappa', 'lambda']
    else:
        chain = [chain, ]
    seqs = [s for s in seqs if s['chain'] in chain]
    cdr3s = [s['cdr3_len'] for s in seqs if s['cdr3_len'] > 0 and s['cdr3_len'] <= max_len]
    x, y = _aggregate(cdr3s, keys=list(range(1, max_len + 1)))
    color = color if color is not None else sns.hls_palette(7)[4]
    x_title = 'CDR3 Length (AA)'
    y_title = 'Frequency (%)'
    size = ((max_len / 5.0) + 1, 4)
    barplot([str(i) for i in x], y,
            color,
            fig_file=fig_file,
            xlabel=x_title,
            ylabel=y_title,
            grid=True,
            size=size,
            xfontsize=7)


def vj_heatmap(seqs, fig_file=None, species='human', chain='heavy'):
    from .base import heatmap
    seqs = [s for s in seqs if s['chain'] == chain]
    vj_data = _group_by_vj(seqs, species, chain)
    vj_df = pd.DataFrame(vj_data)
    heatmap(vj_df.transpose(), fig_file=fig_file)



#==============================
#        PLOT UTILS
#==============================


def _aggregate(data, norm=True, sort_by='value', keys=None):
    '''
    Counts the number of occurances of each item in 'data'.

    Inputs
    data: a list of values.
    norm: normalize the resulting counts (as percent)
    sort_by: how to sort the retured data. Options are 'value' and 'count'.

    Output
    a non-redundant list of values (from 'data') and a list of counts.
    '''
    if keys:
        vdict = {k: 0 for k in keys}
        for d in data:
            if d in keys:
                vdict[d] += 1
    else:
        vdict = {}
        for d in data:
            vdict[d] = vdict[d] + 1 if d in vdict else 1
    vals = [(k, v) for k, v in vdict.items()]
    if sort_by == 'value':
        vals.sort(key=lambda x: x[0])
    else:
        vals.sort(key=lambda x: x[1])
    xs = [v[0] for v in vals]
    if norm:
        raw_y = [v[1] for v in vals]
        total_y = sum(raw_y)
        ys = [100. * y / total_y for y in raw_y]
    else:
        ys = [v[1] for v in vals]
    return xs, ys


def _get_germline_plot_colors(data):
    fams = [d.split('-')[0].replace('/OR', '') for d in data]
    nr_fams = sorted(list(set(fams)))
    num_colors = len(nr_fams)
    rgbs = sns.hls_palette(num_colors)
    rgb_dict = {i[0]: i[1] for i in zip(nr_fams, rgbs)}
    return [rgb_dict[f] for f in fams]


def _group_by_vj(data, species, chain):
    # from abtools.utils.germlines import germlines
    vs = list(set([g.name.split('*')[0] for g in get_germlines(species, 'V', chain=chain)]))
    js = list(set([g.name.split('*')[0] for g in get_germlines(species, 'J', chain=chain)]))
    vj = {}
    for v in vs:
        vj[v] = {}
        for j in js:
            vj[v][j] = 0
    for d in data:
        v = d['v_gene']['gene']
        j = d['j_gene']['gene'].rstrip('P')
        if v[:3] != j[:3]:
            continue
        if v not in vj:
            continue
        if j not in vj[v]:
            continue
        vj[v][j] = vj[v][j] + 1 if j in vj[v] else 1
    total = len(data)
    for v in list(vj.keys()):
        for j in list(vj[v].keys()):
            vj[v][j] = 100. * vj[v][j] / total
    return vj




#!/usr/bin/env python
# filename: bar.py


#
# Copyright (c) 2022 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from collections import Counter
import itertools
import os
import re
import sys

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.lines import Line2D
from matplotlib.patches import Patch
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

import scanpy as sc

import seaborn as sns

from anndata import AnnData

from natsort import natsorted

from abutils.utils.color import get_cmap
from abutils.utils.utilities import nested_dict_lookup




def bar(
    x=None,
    y=None, 
    hue=None, 
    data=None,
    order=None, 
    hue_order=None, 
    palette=None,
    color=None,
    alt_color='#D3D3D3',
    normalize=False, 
    highlight=None, 
    highlight_color=None,  
    orientation='vertical',
    plot_kwargs=None,
    legend_kwargs=None,
    hide_legend=False,
    xlabel=None,
    ylabel=None,
    xlabel_fontsize=16,
    ylabel_fontsize=16,
    xtick_labelsize=14,
    ytick_labelsize=14,
    xtick_labelrotation=0,
    ytick_labelrotation=0,
    show=False,
    figsize=None,
    figfile=None,):
    '''
    Produces a bar plot of categorical data. For data with distinct batches, a stacked 
    bar plot will be constructed.

    Parameters
    ----------

    x : str or list, optional  
        Name of a column in `data` or an iterable of values to be plotted on the 
        x-axis. Required if `orientation` is ``'horizontal'``.

    y : str or list, optional  
        Name of a column in `data` or an iterable of values to be plotted on the 
        y-axis. Required if `orientation` is ``'horizontal'``.

    hue : str, optional  
        Name of a column in `data` or an iterable of hue categories to be used to 
        group data into stacked bars. If not provided, an un-stacked bar plot is created.  

    data : pandas.DataFrame, optional  
        A ``DataFrame`` object containing the input data. If provided, `x` and/or `y` should 
        be column names in `data`.
        
    order : iterable object, optional  
        List of `x` or `y` categories in the order they should be plotted. If `order` contains a 
        subset of all categories found in `x` or `y`, only the supplied categories will be plotted.  
        If not provided, categories will be plotted in ``natsort.natsorted()`` order.
        
    hue_order : iterable object, optional  
        List of `hue` categories in the order they should be plotted. If `hue_order` contains a 
        subset of all categories found in `hue`, only the supplied categories will be plotted.  
        If not provided, `hue` categories will be plotted in ``natsort.natsorted()`` order.
        
    palette : dict, optional  
        Dictionary mapping `hue`, `x` or `y` names to colors. If if keys in `palette` match 
        more than one category, `hue` categories take priority. If `palette` is not provided, 
        bars are colored using `color` (if `hue` is ``None``) or a palette is generated 
        automatically using ``sns.hls_palette()``.   

    color : str or iterable, optional  
        Single color to be used for the bar plot. If not provided, the first color in the 
        default ``Seaborn`` color palette will be used. If `highlight` is provided but 
        `highlight_color` is not, `color` will be used to color highlighted bars.  
        
    alt_color : str or iterable, default='#D3D3D3'  
        Alternate color for the bar plot. Used to color categories not provided in `palette` 
        or to color categories not present in `highlight`.  

    orientation : str, optional
        Orientation of the plot. Options are ``'vertical'`` or ``'horizontal'``. Default is 
        ``'vertical'``.
        
    normalize : bool, default=False  
        If ``True``, normalized frequencies are plotted instead of raw counts. If multiple `hue`
        categories are present, data will be normalized such that all 
        bars extend from [0,1] and each stacked bar is sized according to the `hue`'s  fraction. 
        If `hue` is not provided or there is only one `hue` category, the entire 
        dataset is normalized.
        
    highlight : iterable, optional
        List of `x` or `hue` categories to be highlighted. If `highlight_color` is provided, 
        categories in `highlight` will use `highlight_color` and all others will use `alt_color`. 
        If `highlight_color` is not provided, `palette` will be used. If both `highlight_color` 
        and `palette` are not provided, `color` will be used. 
        
    highlight_color : str or iterable, optional  
        Color to be used for categories in `highlight`. If  

    plot_kwargs : dict, optional  
        Dictionary containing keyword arguments that will be passed to ``pyplot.bar()``.

    legend_kwargs : dict, optional  
        Dictionary containing keyword arguments that will be passed to ``ax.legend()``.

    hide_legend : bool, default=False  
        By default, a plot legend will be shown if multiple batches are plotted. If ``True``, 
        the legend will not be shown.  
        
    xlabel : str, optional  
        Text for the X-axis label. 

    ylabel : str, optional  
        Text for the Y-axis label.  
        
    xlabel_fontsize : int or float, default=16  
        Fontsize for the X-axis label text.

    ylabel_fontsize : int or float, default=16  
        Fontsize for the Y-axis label text.

    xtick_labelsize : int or float, default=14  
        Fontsize for the X-axis tick labels.  

    ytick_labelsize : int or float, default=14  
        Fontsize for the Y-axis tick labels.  

    xtick_labelrotation : int or float, default=0  
        Rotation of the X-axis tick labels.  
    
    ytick_labelrotation : int or float, default=0  
        Rotation of the Y-axis tick labels. 

    show :bool, default=False  
        If ``True``, plot is shown and the plot ``Axes`` object is not returned. Default
        is ``False``, which does not call ``pyplot.show()`` and returns the ``Axes`` object.

    figsize : iterable object, default=[6, 4]  
        List containing the figure size (as ``[x-dimension, y-dimension]``) in inches. 

    figfile : str, optional  
        Path at which to save the figure file. If not provided, the figure is not saved
        and is either shown (if `show` is ``True``) or the ``Axes`` object is returned.  
    '''
    # process input data
    if orientation == 'horizontal':
        if y is not None:
            x, y = y, x
    if data is None:
        _data = {}
        if x is not None:
            _data['x'] = x
            x = 'x'
        if y is not None:
            _data['y'] = y
            y = 'y'
        if hue is not None:
            _data['hue'] = hue
            hue = 'hue'
        data = pd.DataFrame(_data)
    
    # figure size
    if figsize is None:
        if orientation == 'horizontal':
            figsize = [4, 6]
        else:
            figsize = [6, 4]

    if order is None:
        x_vals = natsorted(data[x].unique())
    else:
        x_vals = order
    
    # process hue, if provided
    if hue is not None:
        hue_vals = data[hue]
        hue_order = (
            hue_order
            if hue_order is not None
            else natsorted(set(hue_vals))
        )
        hue_batches = [data[[_hue_val == h for _hue_val in hue_vals]] for h in hue_order]
    else:
        hue_order = [
            None,
        ]
        hue_batches = [
            data,
        ]
        
    # process batches
    batch_data = []
    for batch in hue_batches:
        if y is not None:
            y_dict = {}
            for _x, _y in zip(batch[x], batch[y]):
                if _x in y_dict:
                    y_dict[_x] += float(_y)
                else:
                    y_dict[_x] = float(_y)
            batch_data.append(y_dict)
        else:
            y_dict = Counter(batch[x])
            batch_data.append(y_dict)
    if normalize:
        if len(batch_data) > 1:
            ytots = {xval: sum([b[xval] for b in batch_data]) for xval in x_vals}
            for y_dict in batch_data:
                for xval in x_vals:
                    yval = y_dict.get(xval, 0)
                    y_dict[xval] = yval / ytots[xval]
        else:
            for ydict in batch_data:
                tot = sum(ydict.values())
                for xval in x_vals:
                    yval = y_dict.get(xval, 0)
                    y_dict[xval] = yval / tot
        
    # colors
    if palette is None:
        if len(hue_batches) > 1:
            palette = {h: c for h, c in zip(hue_order, sns.hls_palette(len(hue_order)))}
        else:
            palette = {}
        # palette is a dict assigning colors to x and/or hue categories
        # if both are present, hue takes priority
    colors = []
    color = color if color is not None else sns.color_palette()[0]
    for _hue, batch in zip(hue_order, hue_batches):
        _colors = []
        for _x in x_vals:
            if highlight is not None:
                if _x in highlight:
                    if highlight_color is not None:
                        _colors.append(highlight_color)
                    else:
                        _colors.append(palette.get(_hue, palette.get(_x, color)))
                else:
                    _colors.append(alt_color)
            else:
                _colors.append(palette.get(_hue, palette.get(_x, alt_color)))
        colors.append(_colors)
            
    # plot kwargs
    if orientation == 'horizontal':
        default_plot_kwargs = {"height": 0.8, "linewidth": 1.5, "edgecolor": "w"}
    else:
        default_plot_kwargs = {"width": 0.8, "linewidth": 1.5, "edgecolor": "w"}
    if plot_kwargs is not None:
        default_plot_kwargs.update(plot_kwargs)
    plot_kwargs = default_plot_kwargs

    # legend kwargs
    default_legend_kwargs = {"frameon": True, "loc": "best", "fontsize": 12}
    if legend_kwargs is not None:
        default_legend_kwargs.update(legend_kwargs)
    legend_kwargs = default_legend_kwargs
    
    # make the plot
    plt.figure(figsize=figsize)
    bottom = np.zeros(len(x_vals))
    for h, d, c in zip(hue_order, batch_data, colors):
        y_vals = np.asarray([d.get(_x, 0) for _x in x_vals])
        if orientation == 'horizontal':
            plt.barh(x_vals, y_vals, left=bottom, color=c, label=h, **plot_kwargs)
        else:
            plt.bar(x_vals, y_vals, bottom=bottom, color=c, label=h, **plot_kwargs)
        bottom += y_vals
        
    # style the plot
    ax = plt.gca()
    if orientation == 'horizontal':
        if xlabel is None:
            xlabel = "Frequency (%)" if normalize else "Count"
    else:
        if ylabel is None:
            ylabel = "Frequency (%)" if normalize else "Count"
    ax.set_xlabel(xlabel, fontsize=xlabel_fontsize)
    ax.set_ylabel(ylabel, fontsize=ylabel_fontsize)
    ax.tick_params(
        axis="x", labelsize=xtick_labelsize, labelrotation=xtick_labelrotation
    )
    ax.tick_params(axis="y", labelsize=ytick_labelsize, labelrotation=ytick_labelrotation)
    for s in ["left", "right", "top"]:
        ax.spines[s].set_visible(False)
        
    if orientation == 'horizontal':
        ax.set_ylim([-0.75, len(x_vals) - 0.25])
    else:
        ax.set_xlim([-0.75, len(x_vals) - 0.25])

    # legend
    if len(hue_batches) > 1 and not hide_legend:
        if orientation == 'horizontal':
            leg_kwargs = {'loc': 'center left', 
                          'bbox_to_anchor': [1.01, 0.5]}
            leg_kwargs.update(legend_kwargs)
        else:
            leg_kwargs = legend_kwargs
        ax.legend(**leg_kwargs)
    if hide_legend or palette is None:
        ax.get_legend().remove()

    # save, show or return the ax
    if figfile is not None:
        plt.tight_layout()
        plt.savefig(figfile)
    elif show:
        plt.show()
    else:
        return ax









#!/usr/bin/env python
# filename: io.py


#
# Copyright (c) 2020 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import sys

from .core.sequence import read_csv, read_fasta, read_json, from_mongodb

from .utils.convert import abi_to_fasta
from .utils.pipeline import list_files, make_dir



def read_sequences(file=None, format='tabular', sep='\t', fields=None, match=None,
                   id_key='sequence_id', sequence_key='sequence',
                   db=None, collection=None, mongodb_kwargs=None):
    '''
    Reads sequence data from a file and returns ``abutils.Seqeunce`` objects.

    Args:
    -----

    file (str): path to a file containing sequence data in any of the supported formats.

    format (str): format of the sequence file. Supported formats are: ``'tabular'``, ``'fasta'``,
        ``'json'`` and ``'mongodb'``. Default is ``'tabular'``.

    sep (str): character used to separate fields in ``'tabular'`` input files. This option is
        only used when ``format`` is ``'tabular'``. Default is ``'\t'``, which conforms with the 
        default format for AIRR-compatible sequence annotation.

    id_key (str): name of the field containing the sequence ID. Default is ``'sequence_id'``.

    sequence_key (str): name of the field containing the sequence. Default is ``'sequence'``.

    db (str): mongodb database to query for sequence information. Required if ``format`` is ``'mongodb'``.

    collection (str): mongodb collection to query for sequence information. Required if ``format`` is ``'mongodb'``.

    mongodb_kwargs (dict): dictionary containing additional keyword arguments that will be passed to 
        ``abutils.io.from_mongodb``.
    

    Returns:
    --------

    A list of ``abutils.Sequence`` objects.
    '''
    format = format.lower()
    if format == 'json':
        return read_json(file, id_key=id_key, sequence_key=sequence_key, fields=fields, match=match)
    elif format == 'fasta':
        return read_fasta(file)
    elif format == 'tabular':
        return read_csv(file, delimiter=sep, id_key=id_key, sequence_key=sequence_key, fields=fields, match=match)
    elif format == 'mongodb':
        if any([db is None, collection is None]):
            error = f'ERROR: db and collection are required arguments if the data type is "mongodb".'
            print(error)
            sys.exit()
        return from_mongodb(db, collection, **mongodb_kwargs)
    else:
        error = f'ERROR: format type "{format}"" is not supported. '
        error += f'supported file types are "fasta", "json" and "tabular".'
        print(error)
        sys.exit()











#!/usr/bin/env python
# filename: ssh_tunnel.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


# Borrowed heavily from Paramiko:
# https://github.com/paramiko/paramiko/blob/master/demos/forward.py


import getpass
import os
import socket
import select
try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer

import argparse
import sys

import paramiko

SSH_PORT = 22
DEFAULT_PORT = 27017


class ForwardServer(socketserver.ThreadingTCPServer):
    daemon_threads = True
    allow_reuse_address = True


class Handler(socketserver.BaseRequestHandler):
    def handle(self):
        try:
            chan = self.ssh_transport.open_channel(
                'direct-tcpip',
                (self.chain_host, self.chain_port),
                self.request.getpeername())
        except Exception as e:
            print('Incoming request to {}:{} failed: {}'.format(
                self.chain_host,
                self.chain_port,
                repr(e)))
            return
        if chan is None:
            print('Incoming request to {}:{} was rejected by the SSH server.'.format(
                self.chain_host,
                self.chain_port))
            return

        print('Connected!  Tunnel open {} -> {} -> {}'.format(
            self.request.getpeername(),
            chan.getpeername(),
            (self.chain_host, self.chain_port)))
        while True:
            r, w, x = select.select([self.request, chan], [], [])
            if self.request in r:
                data = self.request.recv(1024)
                if len(data) == 0:
                    break
                chan.send(data)
            if chan in r:
                data = chan.recv(1024)
                if len(data) == 0:
                    break
                self.request.send(data)
        peername = self.request.getpeername()
        chan.close()
        self.request.close()
        print('Tunnel closed from {}'.format(peername,))


def forward_tunnel(local_port, remote_host, remote_port, transport):
    # this is convoluted, but necessary to configure things for the Handler
    # object.  (SocketServer doesn't give Handlers any way to access the outer
    # server.)
    class SubHander (Handler):
        chain_host = remote_host
        chain_port = remote_port
        ssh_transport = transport
    ForwardServer(('', local_port), SubHander).serve_forever()


def get_host_port(spec, default_port):
    "parse 'hostname:22' into a host and port, with the port optional"
    args = (spec.split(':', 1) + [default_port])[:2]
    args[1] = int(args[1])
    return args[0], args[1]


HELP = """\
Set up a forward tunnel across an SSH server, using paramiko. A local port
(given with -p) is forwarded across an SSH session to an address:port from
the SSH server. This is similar to the openssh -L option.
"""


def parse_arguments():
    parser = argparse.ArgumentParser(usage='usage: ssh_tunnel [options] <ssh-server>[:<server-port>]',
                            version='ssh_tunnel', description=HELP)
    parser.add_argument('ssh_server', nargs=1,
                        help='SSH server, as <ssh-server>[:<ssh-port>]')
    parser.add_argument('-p', '--local-port', action='store', type=int, dest='port',
                        default=DEFAULT_PORT,
                        help='local port to forward (default: {})'.format(DEFAULT_PORT))
    parser.add_argument('-u', '--user', action='store', type=str, dest='user',
                        default=getpass.getuser(),
                        help='username for SSH authentication (default: {})'.format(getpass.getuser()))
    parser.add_argument('-K', '--key', action='store', type=str, dest='keyfile',
                        default=None,
                        help='private key file to use for SSH authentication')
    parser.add_argument('--no-key', action='store_false', dest='look_for_keys', default=True,
                        help="don't look for or use a private key file")
    parser.add_argument('-P', '--password', dest='readpass', default=False, action='store_true',
                        help='Use a password for SSH.')
    parser.add_argument('-r', '--remote', action='store', required=True, type=str, dest='remote',
                        default=None, metavar='host:port',
                        help='remote host and port to forward to')
    args = parser.parse_args()

    if args.remote is None:
        parser.error('Remote address required (-r).')

    server_host, server_port = get_host_port(args.ssh_server[0], SSH_PORT)
    remote_host, remote_port = get_host_port(args.remote, args.port)
    return args, (server_host, server_port), (remote_host, remote_port)


def main():
    args, server, remote = parse_arguments()
    if args.readpass:
        password = getpass.getpass('Enter SSH password: ')

    client = paramiko.SSHClient()
    client.load_system_host_keys()
    client.set_missing_host_key_policy(paramiko.WarningPolicy())

    print('Connecting to ssh host {}:{} ...'.format(server[0], server[1]))
    try:
        client.connect(server[0], server[1], username=args.user, key_filename=args.keyfile,
                       look_for_keys=args.look_for_keys, password=password)
    except Exception as e:
        print('*** Failed to connect to {}:{}: {}'.format(server[0], server[1], e))
        sys.exit(1)

    print('Now forwarding port {} to {}:{} ...'.format(args.port, remote[0], remote[1]))
    try:
        forward_tunnel(args.port, remote[0], remote[1], client.get_transport())
    except KeyboardInterrupt:
        print('C-c: Port forwarding stopped.')
        sys.exit(0)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# filename: sequence.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

from collections import OrderedDict
import csv
import json
import operator
import sys
import uuid

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

import pandas as pd

from ..utils.codons import codon_lookup
from ..utils.utilities import nested_dict_lookup

if sys.version_info[0] > 2:
    STR_TYPES = [str, ]
    from functools import reduce
else:
    STR_TYPES = [str, unicode]


class Sequence(object):
    """
    Container for biological (RNA and DNA) sequences.

    ``seq`` can be one of several things:

        1) a raw sequence, as a string

        2) an iterable, formatted as ``[seq_id, sequence]``

        3) a dict, containing at least the sequence ID and a
           sequence. Alternate ``id_key`` and ``seq_key``
           can be provided at instantiation.

        4) a Biopython ``SeqRecord`` object

        5) an abutils ``Sequence`` object

    If ``seq`` is provided as a string, the sequence ID can optionally be
    provided via the ``id`` keyword argument.  If ``seq`` is a string and ``id`` is not provided,
    a random sequence ID will be generated with ``uuid.uuid4()``.

    Quality scores can be supplied with ``qual`` or as part of a ``SeqRecord`` object.
    If providing both a SeqRecord object with quality scores and quality scores
    via ``qual``, the ``qual`` scores will override the SeqRecord quality scores.

    If ``seq`` is a dictionary, typically the result of a MongoDB query, the dictionary
    can be accessed directly from the ``Sequence`` instance (via the ``annotations`` property). To retrive the value
    for ``'junc_aa'`` in the instantiating dictionary, you would simply::

        s = Sequence(dict)
        junc = s['junc_aa']

    If ``seq`` is a dictionary, an optional ``id_key`` and ``seq_key`` can be provided,
    which tells the ``Sequence`` object which field to use to populate ``Sequence.id`` and
    ``Sequence.sequence``. Defaults for both ``id_key`` and ``seq_key`` are ``None``, which 
    results in abutils trying to determine the appropriate key. For ``id_key``, the following 
    keys are tried: ``['seq_id', 'sequence_id']``. For ``seq_key``, the following keys are tried:
    ``['vdj_nt', 'sequence_nt', 'sequence']``. If none of the attempts are successful, the 
    ``Sequence.id`` or ``Sequence.sequence`` attributes will be ``None``.

    Alternately, the ``__getitem__()`` interface can be used to obtain a slice from the
    ``sequence`` attribute. An example of the distinction::

        d = {'name': 'MySequence', 'sequence': 'ATGC'}
        seq = Sequence(d, id_key='name', seq_key='sequence')

        seq['name']  # 'MySequence'
        seq[:2]  # 'AT'

    If the ``Sequence`` is instantiated with a dictionary, calls to ``__contains__()`` will
    return ``True`` if the supplied item is a key in the dictionary. In non-dict instantiations,
    ``__contains__()`` will look in the ``Sequence.sequence`` field directly (essentially a
    motif search). For example::

        dict_seq = Sequence({'seq_id': 'seq1', 'vdj_nt': 'ACGT'})
        'seq_id' in dict_seq  # TRUE
        'ACG' in dict_seq     # FALSE

        str_seq = Sequence('ACGT', id='seq1')
        'seq_id' in str_seq  # FALSE
        'ACG' in str_seq     # TRUE

    .. note::

        When comparing ``Sequence`` objects, they are comsidered equal only if their
        sequences and IDs are identical. This means that two ``Sequence`` objects
        with identical sequences but without user-supplied IDs won't be equal,
        because their IDs will have been randomly generated.
    """
    def __init__(self, seq, id=None, qual=None, id_key=None, seq_key=None):
        super(Sequence, self).__init__()
        self._input_sequence = None
        self._input_id = id
        self._input_qual = qual
        self.id = None
        self.sequence = None
        self.qual = None
        self._annotations = None
        self._fasta = None
        self._fastq = None
        self._reverse_complement = None
        self._strand = None
        self.id_key = id_key
        self.seq_key = seq_key
        self.description = None

        self._process_input(seq, id, qual)


    def __len__(self):
        '''
        int: Returns the length of the ``sequence`` attribute
        '''
        return len(self.sequence)

    def __iter__(self):
        '''
        iter: Returns an iterator over the ``sequence`` attribute
        '''
        return iter(self.sequence)

    def __reversed__(self):
        '''
        str: Returns the reverse of the ``sequence`` attribute
        '''
        return ''.join(reversed(self.sequence))

    def __contains__(self, item):
        '''
        bool: If the instance was initialzed with a dictonary (which means
            the ``annotations`` attribute is not empty), ``__contains__(key)``
            will return ``True`` if ``key`` is in ``annotations.keys()``. If ``annotations``
            is an empty dict, indicating instantiation without a dictionary,
            ``__contains__(motif)`` will return True if ``motif`` is in the
            ``sequence attribute.

        '''
        if self.annotations:
            return item in self.annotations.keys()
        return item in self.sequence

    def __getitem__(self, key):
        if type(key) == slice:
            return self.sequence[key]
        elif key in self.annotations.keys():
            return self.annotations.get(key, None)
        elif type(key) == int:
            return self.sequence[key]
        return None

    def __setitem__(self, key, val):
        self.annotations[key] = val

    def __eq__(self, other):
        if all([hasattr(other, 'sequence'), hasattr(other, 'id')]):
            return all([self.sequence == other.sequence, self.id == other.id])
        return False

    
    @classmethod
    def from_json(cls, data):
        if isinstance(data, str):
            data = json.loads(data)
        s = cls(None)
        s.__dict__ = data
        return s


    @property
    def fasta(self):
        '''
        str: Returns the sequence, as a FASTA-formatted string

        Note: The FASTA string is built using ``Sequence.id`` and ``Sequence.sequence``.
        '''
        if not self._fasta:
            self._fasta = '>{}\n{}'.format(self.id, self.sequence)
        return self._fasta

    @property
    def fastq(self):
        '''
        str: Returns the sequence, as a FASTQ-formatted string

        If ``Sequence.qual`` is ``None``, then ``None`` will be returned instead of a
        FASTQ string
        '''
        if self.qual is None:
            self._fastq = None
        else:
            if self._fastq is None:
                self._fastq = '@{}\n{}\n+\n{}'.format(self.id, self.sequence, self.qual)
        return self._fastq

    @property
    def reverse_complement(self):
        '''
        str: Returns the reverse complement of ``Sequence.sequence``.
        '''
        if self._reverse_complement is None:
            self._reverse_complement = self._get_reverse_complement()
        return self._reverse_complement

    @property
    def annotations(self):
        if self._annotations is None:
            self._annotations = {}
        return self._annotations

    @annotations.setter
    def annotations(self, val):
        self._annotations = val

    @property
    def strand(self):
        if self._strand is None:
            self._strand = 'plus'
        return self._strand

    @strand.setter
    def strand(self, strand):
        self._strand = strand


    def translate(self, sequence_key=None, frame=1):
        return translate(self, sequence_key, frame)


    def as_fasta(self, name_field=None, seq_field=None):
        name = None
        sequence = None
        if name_field is not None:
            name = self.annotations[name_field]
        if seq_field is not None:
            sequence = self.annotations[seq_field]
        if name is None:
            name = self.id
        if sequence is None:
            sequence = self.sequence
        return '>{}\n{}'.format(name, sequence)


    def region(self, start=0, end=None):
        '''
        Returns a region of ``Sequence.sequence``, in FASTA format.

        If called without kwargs, the entire sequence will be returned.

        Args:

            start (int): Start position of the region to be returned. Default
                is 0.

            end (int): End position of the region to be returned. Negative values
                will function as they do when slicing strings.

        Returns:

            str: A region of ``Sequence.sequence``, in FASTA format
        '''
        if end is None:
            end = len(self.sequence)
        return '>{}\n{}'.format(self.id, self.sequence[start:end])


    def keys(self):
        return self.annotations.keys()


    def values(self):
        return self.annotations.values()


    def get(self, key, default=None):
        return self.annotations.get(key, default)


    def _get_reverse_complement(self):
        rc = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A',
              'Y': 'R', 'R': 'Y', 'S': 'S', 'W': 'W',
              'K': 'M', 'M': 'K', 'B': 'V', 'D': 'H',
              'H': 'D', 'V': 'B', 'N': 'N'}
        return ''.join([rc.get(res, res) for res in self.sequence[::-1]])


    def _process_input(self, seq, id, qual):
        if type(seq) in STR_TYPES:
            if id is None:
                id = uuid.uuid4()
            self.sequence = str(seq).upper()
            self.id = id
            self.qual = qual
            self._input_sequence = self.sequence
        elif type(seq) == Sequence:
            self.id = id if id is not None else seq.id
            self.sequence = seq.sequence
            self.description = seq.description
            self.qual = seq.qual
            self._input_sequence = self.sequence
            self._annotations = seq._annotations
        elif type(seq) in [list, tuple]:
            self.id = str(seq[0])
            self.sequence = str(seq[1]).upper()
            self.qual = qual
            self._input_sequence = self.sequence
        elif type(seq) == SeqRecord:
            if qual is None:
                if 'phred_quality' in seq.letter_annotations:
                    qual = seq.letter_annotations['phred_quality']
                elif 'solexa_quality' in seq.letter_annotations:
                    qual = seq.letter_annotations['solexa_quality']
            self.id = id if id is not None else str(seq.id)
            self.description = str(seq.description)
            self.sequence = str(seq.seq).upper()
            self.qual = qual
            self._input_sequence = self.sequence
        elif type(seq) in [dict, OrderedDict]:
            if self.id_key is None:
                id_options = ['seq_id', 'sequence_id']
                if any([k in seq for k in id_options]):
                    self.id_key = [k for k in id_options if k in seq][0]
            if self.seq_key is None:
                seq_options = ['vdj_nt', 'sequence_nt', 'sequence']
                if any([k in seq for k in seq_options]):
                    self.seq_key = [k for k in seq_options if k in seq][0]
            self.id = id if id is not None else seq.get(self.id_key, None)
            self.sequence = seq.get(self.seq_key, None)
            self.qual = qual
            self._input_sequence = self.sequence
            self._annotations = seq



def translate(sequence, sequence_key=None, frame=1):
    if sequence_key is not None:
        seq = Sequence(nested_dict_lookup(sequence.annotations, sequence_key.split('.')))
    else:
        seq = Sequence(sequence)
    if seq is None:
        return None
    start = (frame % 3) - 1
    end = len(seq) - (len(seq[start:]) % 3)
    seq = seq[start:end]
    translated = ''
    for i in range(0, len(seq), 3):
        codon = seq[i:i+3]
        if all([c == '-' for c in codon]):
            aa = '-'
        else:
            aa = codon_lookup.get(codon, 'X')
        translated += aa
    return translated



def read_json(json_file, match=None, fields=None, id_key='seq_id', sequence_key='sequence'):
    if match is None:
        match = {}
    if fields is not None:
        if id_key not in fields:
            fields.append(id_key)
        if sequence_key not in fields:
            fields.append(sequence_key)
    sequences = []
    with open(json_file, 'r') as f:
        for line in f:
            if not line.strip():
                continue
            j = json.loads(line.strip())
            try:
                if all([nested_dict_lookup(j, k.split('.')) == v for k, v in match.items()]):
                    if fields is not None:
                        _fields = [f for f in fields if f in j]
                        j = {f: j[f] for f in _fields}
                    sequences.append(Sequence(j, id_key=id_key, seq_key=sequence_key))
            except KeyError:
                continue
    return sequences



def read_csv(csv_file, delimiter=',', match=None, fields=None, id_key='sequence_id', sequence_key='sequence'):
    if match is None:
        match = {}
    if fields is not None:
        if id_key not in fields:
            fields.append(id_key)
        if sequence_key not in fields:
            fields.append(sequence_key)
    sequences = []
    df = pd.read_csv(csv_file, delimiter=delimiter)
    for _, r in df.iterrows():
        r = r.to_dict()
        try:
            if all([r[k] == v for k, v in match.items()]):
                if fields is not None:
                    _fields = [f for f in fields if f in r]
                    r = {f: r[f] for f in _fields}
                sequences.append(Sequence(r, id_key=id_key, seq_key=sequence_key))
        except KeyError:
            continue
    return sequences


def read_fasta(fasta_file):
    with open(fasta_file) as f:
        sequences = [Sequence(s) for s in SeqIO.parse(f, 'fasta')]
    return sequences


def from_mongodb(db, collection, match=None, project=None, limit=None):
    if match is None:
        match = {}
    if project is None:
        project = {}
    if limit is not None:
        results = db[collection].find(match, project, limit=limit)
    else:
        results = db[collection].find(match, project)
    return [Sequence(r) for r in results]

            








#!/usr/bin/env python
# filename: utilities.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

import math
import operator
import sys

if sys.version_info[0] > 2:
    STR_TYPES = [str, ]
    from functools import reduce
else:
    STR_TYPES = [str, unicode]




#================================
#       USEFUL SNIPPITS
#================================


def nested_dict_lookup(d, key_list, missing=None):
    '''
    Retrieves a nested value from a dictionary given a list of keys::

        mydict = {'key1a': {'key1b': 'val1b}, 'key2a': {'key2b': {'key2c': 'val2c}}}

        keylist1 = ['key1a', 'key1b']
        nested_dict_lookup(mydict, keylist1)
        >>> val1b

        keylist2 = ['key2a', 'key2b', 'key2c']
        nested_dict_lookup(mydict, keylist2)
        >>> val2c

    Args:

        d (dict): Nested dictionary

        keylist (list(str)): list of strings that correspond to dict keys (in nested order)


    Returns:

        Value from the nested dictionary. If one or more of the nested keys is not present
        in the nested dictionary, `None` will be returned.
    '''
    try:
        return reduce(operator.getitem, key_list, d)
    except (KeyError, TypeError):
        return missing


def chunker(iterable, chunksize):
    r = int(math.ceil(len(iterable) / chunksize))
    for i in range(r):
        start = i * chunksize
        end = (i + 1) * chunksize
        yield iterable[start:end]


#!/usr/bin/env python
# filename: scatter.py


#
# Copyright (c) 2022 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from collections import Counter
import itertools

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib as mpl
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

import seaborn as sns

from natsort import natsorted

from abutils.utils.color import get_cmap



def scatter(
    x=None,
    y=None, 
    hue=None, 
    marker="o",
    data=None,
    hue_order=None, 
    force_categorical_hue=False,
    palette=None,
    color=None,
    cmap=None,
    size=20,
    alpha=0.6,
    highlight_index=None,
    highlight_x=None,
    highlight_y=None,
    highlight_marker="x",
    highlight_size=90,
    highlight_color="k",
    highlight_name=None,
    highlight_alpha=0.9,  
    plot_kwargs=None,
    legend_kwargs=None,
    hide_legend=False,
    xlabel=None,
    ylabel=None,
    title=None,
    title_fontsize=20,
    title_fontweight='normal',
    title_loc='center',
    title_pad=None,
    show_title=False,
    xlabel_fontsize=16,
    ylabel_fontsize=16,
    xtick_labelsize=14,
    ytick_labelsize=14,
    xtick_labelrotation=0,
    ytick_labelrotation=0,
    cbar_width=0.35,
    cbar_height=0.05,
    cbar_loc="lower right",
    cbar_orientation="horizontal",
    cbar_bbox_to_anchor=None,
    cbar_flip_ticks=False,
    cbar_title=None,
    cbar_title_fontsize=12,
    hide_cbar=False,
    equal_axes=True,
    ax=None,
    show=False,
    figsize=None,
    figfile=None,):
    '''
    Produces a scatter plot.

    Parameters
    ----------

    x : str or iterable object  
        Name of a column in `data` or an iterable of values to be plotted on the 
        x-axis. Required.

    y : str or iterable object  
        Name of a column in `data` or an iterable of values to be plotted on the 
        y-axis..

    hue : str or iterable object, optional  
        Name of a column in `data` or an iterable of hue categories to be used to 
        group data into stacked bars. If not provided, an un-stacked bar plot is created.  
        
    marker : str, dict or iterable object, optional  
        Marker style for the scatter plot. Accepts any of the following:
          * a `matplotlib marker`_ string
          * a ``dict`` mapping `hue` categories to a `matplotlib marker`_ string
          * a ``list`` of `matplotlib marker`_ strings, which should be the same 
              length as `x` and `y`.

    data : pandas.DataFrame, optional  
        A ``DataFrame`` object containing the input data. If provided, `x` and/or `y` should 
        be column names in `data`.   
        
    hue_order : iterable object, optional  
        List of `hue` categories in the order they should be plotted. If `hue_order` contains a 
        subset of all categories found in `hue`, only the supplied categories will be plotted.  
        If not provided, `hue` categories will be plotted in ``natsort.natsorted()`` order.  
        
    force_categorical_hue : bool, default=False  
        If ``True``, `hue` data will be treated as categorical, even if the data appear to 
        be continuous. This results in `color` being used to color the points rather than `cmap`.   
        
    palette : dict, optional  
        Dictionary mapping `hue`, `x` or `y` names to colors. If if keys in `palette` match 
        more than one category, `hue` categories take priority. If `palette` is not provided, 
        bars are colored using `color` (if `hue` is ``None``) or a palette is generated 
        automatically using ``sns.hls_palette()``.  

    color : str or iterable object, optional. 
        Color for the plot markers. Can be any of the following:
          * a ``list`` or ``tuple`` containing RGB or RGBA values
          * a color string, either from `Matplotlib's set of named colors`_ or a hex color value
          * the name of a column in `data` that contains color values
          * a ``list`` of colors (either as strings or RGB tuples) to be used for `hue` categories. 
            If `colors` is shorter than the list of hue categories, colors will be reused.  
        
        .. tip::
            If a single RGB or RGBA ``list`` or ``tuple`` is provided and `hue` is also supplied, 
            there may be unexpected results as ``scatter()`` will attempt to map each of the 
            individual RGB(A) values to a hue category.   
        
        Only used if `hue` contains categorical data (`cmap` is used for continuous data). If not 
        provided, the `default Seaborn color palette`_ will be used. 
        
    cmap : str or matplotlib.color.Colormap, default='flare'   
        Colormap to be used for continuous `hue` data.  
        
    size : str or float or iterable object, default=20  
        Size of the scatter points. Either a   

    alpha : float, default=0.6  
        Alpha of the scatter points.  
        
    highlight_index : iterable object, optional  
        An iterable of index names (present in `data.index`) of points to be highlighted on 
        the plot. If provided, `highlight_x` and `highlight_y` are ignored.

    highlight_x : iterable object, optional  
        An iterable of x-values for highlighted points. Also requires `highlight_y`.
        
    highlight_y : iterable object, optional  
        An iterable of y-values for highlighted points. Also requires `highlight_x`.
        
    highlight_marker : str, default='x'  
        Marker style to be used for highlight points. Accepts any `matplotlib marker`_. 

    highlight_size : int, default=90  
        Size of the highlight marker.

    highlight_color : string or list of color values, default='k'
        Color of the highlight points.

    highlight_name : str, optional  
        Name of the highlights, to be used in the plot legend. If not supplied,
        highlight points will not be included in the legend.
        
    highlight_alpha : float, default=0.9  
        Alpha of the highlight points. 

    plot_kwargs : dict, optional  
        Dictionary containing keyword arguments that will be passed to ``pyplot.scatter()``.

    legend_kwargs : dict, optional  
        Dictionary containing keyword arguments that will be passed to ``ax.legend()``.

    hide_legend : bool, default=False  
        By default, a plot legend will be shown if multiple batches are plotted. If ``True``, 
        the legend will not be shown.  
        
    xlabel : str, optional  
        Text for the X-axis label. 

    ylabel : str, optional  
        Text for the Y-axis label.  
        
    xlabel_fontsize : int or float, default=16  
        Fontsize for the X-axis label text.

    ylabel_fontsize : int or float, default=16  
        Fontsize for the Y-axis label text.

    xtick_labelsize : int or float, default=14  
        Fontsize for the X-axis tick labels.  

    ytick_labelsize : int or float, default=14  
        Fontsize for the Y-axis tick labels.  

    xtick_labelrotation : int or float, default=0  
        Rotation of the X-axis tick labels.  
    
    ytick_labelrotation : int or float, default=0  
        Rotation of the Y-axis tick labels. 

    cbar_width : int, default=35  
        Width of the colorbar. Only used for continuous `hue` types.  

    cbar_height : int, default=5  
        Height of the colorbar. Only used for continuous `hue` types.  

    cbar_loc : str or iterable object, default='lower right'  
        Location of the colorbar. Accepts `any valid inset_axes() location`_.

    cbar_orientation : str, default='horizontal'  
        Orientation of the colorbar. Options are ``'horizontal'`` and ``'vertical'``.

    cbar_bbox_to_anchor : list or tuple, optional
        bbox_to_anchor for the colorbar. Used in combination with `cbar_loc` to provide 
        fine-grained positioning of the colorbar.

    cbar_flip_ticks : bool, default=False  
        Flips the position of colorbar ticks. Ticks default to the bottom if `cbar_orientation` 
        is  ``'horizontal'`` and the left if  `cbar_orientation` is ``'vertical'``.  

    cbar_title : str, optional  
        Colorbar title. If not provided, `hue` is used.  

    cbar_title_fontsize : int or float, default=12  
        Fontsize for the colorbar title.  
        
    hide_cbar : bool, default=False. 
        If ``True``, the color bar will be hidden on plots with continuous `hue` values.  
        
    equal_axes : bool, default=True
        If ```True```, the the limits of the x- and y-axis will be equal.

    ax : mpl.axes.Axes, default=None
        Pre-existing axes for the plot. If not provided, a new axes will be created.

    show :bool, default=False  
        If ``True``, plot is shown and the plot ``Axes`` object is not returned. Default
        is ``False``, which does not call ``pyplot.show()`` and returns the ``Axes`` object.

    figsize : iterable object, default=[6, 4]  
        List containing the figure size (as ``[x-dimension, y-dimension]``) in inches. 

    figfile : str, optional  
        Path at which to save the figure file. If not provided, the figure is not saved
        and is either shown (if `show` is ``True``) or the ``Axes`` object is returned.  
        
        
    .. _matplotlib marker: 
        https://matplotlib.org/stable/api/markers_api.html  

    .. _Matplotlib's set of named colors:
        https://matplotlib.org/stable/gallery/color/named_colors.html
        
    .. _default Seaborn color palette: 
        https://seaborn.pydata.org/generated/seaborn.color_palette.html  

    .. _any valid inset_axes() location: 
        https://matplotlib.org/stable/api/_as_gen/mpl_toolkits.axes_grid1.inset_locator.inset_axes.html
        
    '''
    # process input data
    if data is None:
        _data = {}
        _data['x'] = x
        x = 'x'
        _data['y'] = y
        y = 'y'
        if hue is not None:
            _data['hue'] = hue
            hue = 'hue'
        df = pd.DataFrame(_data)
    else:
        df = data.copy()
    
    # figure size
    if figsize is None:
        figsize = [6, 6]
        
    # hue and color
    continuous_hue = False
    if hue is not None:
        if all([isinstance(h, float) for h in df[hue]]) and not force_categorical_hue:
            continuous_hue = True
            hue_order = []
            if cmap is None:
                cmap = sns.color_palette("flare", as_cmap=True)
            else:
                cmap = plt.get_cmap(cmap)
            min_hue = max(0, df[hue].min())
            max_hue = np.ceil(df[hue].max())
            df["color"] = [cmap((h - min_hue) / (max_hue - min_hue)) for h in df[hue]]
        else:
            if hue_order is None:
                hue_order = natsorted(list(set(df[hue])))
            if palette is not None:
                missing_color = color if color is not None else 'lightgrey'
                hue_dict = {h: palette.get(h, missing_color) for h in hue_order}
            else:
                n_colors = max(1, len(hue_order))
                if color is None:
                    color = sns.color_palette(n_colors=n_colors)
                if len(color) < n_colors:
                    color = itertools.cycle(color)
                hue_dict = {h: c for h, c in zip(hue_order, color)}
            df["color"] = [hue_dict[h] for h in df[hue]]
    else:
        hue_order = []
        if isinstance(color, str) and color in df.columns:
            df["color"] = df[color]
        elif color is not None:
            df["color"] = [color] * df.shape[0]
        else:
            df["color"] = [sns.color_palette()[0]] * df.shape[0]
            
    # markers
    # TODO
    
            
    # plot kwargs
    default_plot_kwargs = {'linewidths': 0}
    if plot_kwargs is not None:
        default_plot_kwargs.update(plot_kwargs)
    plot_kwargs = default_plot_kwargs
    
    
    # scatterplot
    if ax is None:
        plt.figure(figsize=figsize)
        ax = plt.gca()
    if hue_order:
        for h in hue_order[::-1]:
            d = df[df[hue] == h]
            ax.scatter(
                d[x],
                d[y],
                c=d["color"],
                s=size,
                marker=marker,
                alpha=alpha,
                label=h,
                **plot_kwargs,
            )
    else:
        ax.scatter(
            df[x],
            df[y],
            c=df["color"],
            s=size,
            marker=marker,
            alpha=alpha,
            **plot_kwargs,
        )

    # highlighted points
    highlight = any(
        [
            highlight_index is not None,
            all([highlight_x is not None, highlight_y is not None]),
        ]
    )
    if highlight:
        if highlight_index is not None:
            hi_index = [h for h in highlight_index if h in df.index.values]
            hi_df = df.loc[hi_index]
            highlight_x = hi_df[x]
            highlight_y = hi_df[y]
        plt.scatter(
            highlight_x,
            highlight_y,
            zorder=10,
            s=highlight_size,
            c=highlight_color,
            alpha=highlight_alpha,
            marker=highlight_marker,
            label=highlight_name,
        )
    
    # legend
    if not continuous_hue:
        if hue is not None and not hide_legend:
            default_legend_kwargs = {"frameon": True, "loc": "best", "fontsize": 12}
            if legend_kwargs is not None:
                default_legend_kwargs.update(legend_kwargs)
            legend_kwargs = default_legend_kwargs
            ax.legend(**legend_kwargs)
    
    # colorbar
    elif not hide_cbar:
        if cbar_orientation == 'horizontal':
            width = max([cbar_width, cbar_height])
            height = min([cbar_width, cbar_height])
        else:
            width = min([cbar_width, cbar_height])
            height = max([cbar_width, cbar_height])
        cbar_bounds = get_inset_axes_bounds(
            cbar_loc,
            cbar_bbox_to_anchor,
            width,
            height
        )
        cbax = ax.inset_axes(cbar_bounds)

        max_hue = np.ceil(df[hue].max())
        min_hue = max(0, df[hue].min())
        norm = mpl.colors.Normalize(vmin=min_hue, vmax=max_hue)
        # ticks = [t for t in np.linspace(min_hue, max_hue, num=4)]
        
        cbar = plt.colorbar(
            mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
            cax=cbax,
            orientation=cbar_orientation,
        )
        if cbar_orientation == "horizontal":
            ticks_position = "bottom" if cbar_flip_ticks else "top"
            cbax.xaxis.set_ticks_position(ticks_position)
            cbax.xaxis.set_label_position(ticks_position)
            cbar.ax.set_xlabel(
                cbar_title,
                fontsize=cbar_title_fontsize,
            )
        else:
            ticks_position = "left" if cbar_flip_ticks else "right"
            cbax.yaxis.set_ticks_position(ticks_position)
            cbax.yaxis.set_label_position(ticks_position)
            cbar.ax.set_ylabel(
                cbar_title,
                fontsize=cbar_title_fontsize,
            )

    # style the plot
    ax.set_xlabel(xlabel if xlabel is not None else x, fontsize=xlabel_fontsize)
    ax.set_ylabel(ylabel if ylabel is not None else y, fontsize=ylabel_fontsize)
    ax.tick_params(
        axis="x", labelsize=xtick_labelsize, labelrotation=xtick_labelrotation
    )
    ax.tick_params(axis="y", labelsize=ytick_labelsize, labelrotation=ytick_labelrotation)

    for spine in ["right", "top"]:
        ax.spines[spine].set_visible(False)
    for spine in ["left", "bottom"]:
        ax.spines[spine].set_position(("outward", 10))

    if equal_axes:
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        axlim = [min([xlim[0], ylim[0]]), max([xlim[1], ylim[1]])]
        ax.set_xlim(axlim)
        ax.set_ylim(axlim)
        
    if show_title and title is not None:
        ax.set_title(
            title,
            loc=title_loc,
            pad=title_pad,
            fontsize=title_fontsize,
            fontweight=title_fontweight,
        )

    # save, show or return the ax
    if figfile is not None:
        plt.tight_layout()
        plt.savefig(figfile)
    elif show:
        plt.show()
    else:
        return ax




def get_inset_axes_bounds(loc, bbox_to_anchor, width, height):
    if bbox_to_anchor is None:
        loc_dict = {
            'upper left': [0, 1-height],
            'upper center': [0.5-width/2, 1-height],
            'upper right': [1-width, 1-height],
            'center left': [0, 0.5-height/2],
            'center': [0.5-width/2, 0.5-height/2],
            'center right': [1-width, 0.5-height/2],
            'lower left': [0, 0],
            'lower center': [0.5-width/2, 0],
            'lower right': [1-width, 0]
        }
        x0, y0 = loc_dict.get(loc, [0, 0])
    else:
        x, y = bbox_to_anchor[:2]
        loc_dict = {
            'upper left': [x, y-height],
            'upper center': [x-width/2, y-height],
            'upper right': [x-width, y-height],
            'center left': [x, y-height/2],
            'center': [x-width/2, y-height/2],
            'center right': [x-width, y-height/2],
            'lower left': [x, y],
            'lower center': [x-width/2, y],
            'lower right': [x-width, y]
        }
        x0, y0 = loc_dict.get(loc, [0, 0])
    return [x0, y0, width, height]
